local slurm = {}

local safer = require("safer")
local util = require("sga.util")
local exec = require("sga.exec")
local schema = require("schema")

local cmds = {
    sbatch = "sbatch ",
    squeue = "scontrol show job ",
    sinfo = "sinfo -o '%n state=%t np=%c \
             status=%a maxcpu=%B tmpdsksize=%d \
             freemem=%e reason=%E feature=%f \
             group=%g gres=%G oversubscribecompres=%h \
             reasontimestamp=%H partitionpriority=%I \
             maxjobtime=%l defaultjobtime=%L availmem=%m \
             preemption=%M cpuload=%O partitionschedulingpriority=%p \
             partition=%P rootonly=%r maxjobsize=%s schedulingweight=%w \
             sockets=%X cores=%Y threads=%Z' -h",
    scancel = "scancel "
}

--- Type of the SGA, returned to the server during registration.
slurm.type = "cluster"

---
-- Execute a new command.
-- @param job The job object: job.data is a writable table for driver data.
-- @param cmd_string The command string
-- @return True if succeded or nil and an error message
function slurm.execute_command(self, job, cmd_string)
    local script_filename = self.config.runtime_data_dir.."/sbatch_"..job.jid..".script"
    local out_filename = self.config.runtime_data_dir.."/sbatch_"..job.jid..".out"
    local err_filename = self.config.runtime_data_dir.."/sbatch_"..job.jid..".err"
    local sudo_cmd = "sudo "
    local as_root = false

    for _, sandbox_path in ipairs(job.sandboxes) do
        local ok, err = self.exec:create_dir(sandbox_path)
        if not ok then
           if not self.exec:is_dir(sandbox_path) then
              return nil, "Failed creating job's sandbox "..sandbox_path
           end
        end
        self.exec:chmod(sandbox_path, "rwxrwxrwx")
    end

    local optional_params = ""
    if self.slurm_init_dir then
        optional_params = optional_params.." -D "..self.slurm_init_dir
    end
    -- O equivalente de queue no slurm é partição!
    if self.slurm_queue then
        optional_params = optional_params.." -p "..self.slurm_queue
    end
    if self.allow_proxy_user and (job.parameters.slurm_user or job.parameters.csbase_command_user_id) then
        local gid_param =  " --gid="
        if self.proxy_user_group then
            gid_param = gid_param..self.proxy_user_group
        else
            gid_param = gid_param.."$(id -g)"
        end
        optional_params = optional_params.." --uid="..(job.parameters.slurm_user or job.parameters.csbase_command_user_id)..gid_param
        as_root = true
    end

    -- Add umask at top of the  script to ensure that all files created during the
    -- job's execution could be modified by the unix group members
    self.exec:write_file(script_filename, "#!/bin/sh\numask 002\n"..cmd_string.."\n")

    local cmd_prefix = as_root and sudo_cmd or ""
    local full_cmd = ("%s%s --parsable -o %s -e %s %s %s"):format(cmd_prefix, cmds.sbatch, out_filename, err_filename, optional_params, script_filename)
    self.logger:debug("[COMMAND] "..full_cmd)
    local slurmjid, stderr = self.exec:run(full_cmd)
    
    if slurmjid ~= "" then
        slurmjid = slurmjid:gsub("\n", "")
        self.logger:debug("Submitted Slurm job: "..slurmjid)
        job.data.script_filename = script_filename
        job.data.out_filename = out_filename
        job.data.err_filename = err_filename
        job.data.slurmjid = slurmjid
        return true
    else
        local err = "Failed submitting job: "..stderr
        return nil, err
    end
end

function slurm.cleanup_job(self, job)
    self.exec:remove_file(job.data.script_filename)
    self.exec:remove_file(job.data.out_filename)
    self.exec:remove_file(job.data.err_filename)

    for _, sandbox_path in ipairs(job.sandboxes) do
        if self.exec:is_dir(sandbox_path) then
           local ok, err = self.exec:remove_dir(sandbox_path)
           if not ok then
              self.logger:error("Failed removing job's sandbox "..sandbox_path)
           end
        end
    end
end

local function parse(input)
   local out = {}
   input = input:gsub("\n"," ")
   for attr, val in input:gmatch("(%S+)=(%S+)") do
      out[attr]=out[attr] or val
   end
   return out
end

local function squeue(self, job)
    local jdata, stderr = self.exec:run(cmds.squeue..job.data.slurmjid)
    if not jdata or jdata == "" then
        return nil, "Failed running job status command"..(stderr and " - "..stderr)
    end
    local data = parse(jdata) --// Função parse pega o output do squeue e faz o parse do state?
    if not (data) then
        return nil, "Failed parsing job status data"
    end
    job.data.squeue_data = data
    return data
end

local slurm_to_sga_state = {
    COMPLETED = "FINISHED",
    COMPLETING = "RUNNING",
    RUNNING = "RUNNING",
    PENDING = "WAITING",
    CONFIGURING = "WAITING",
    FAILED = "FINISHED"
}

local function get_seconds(time)
    local h, m, s = time:match("(%d+):(%d+):(%d+)")
    if not h then return 0 end
    return s + m * 60 + h * 3600
end

function slurm.is_command_done(self, job)
    local data, err = squeue(self, job)
    if err then
        return nil, err
    end
    job.data.squeue_data = data
    if slurm_to_sga_state[data.JobState] == "FINISHED" then
        local donetime = get_seconds(data.RunTime)
        for k, v in util.collect_exec_data(job) do
        end
        return true, donetime, donetime, donetime -- FIXME detailed times
    else
        return false
    end
end

local mem_fact = {b = 1, k = 1024, m = 1024^2, g = 1024^3, t = 1024^4} 
local function get_mem(minfo)
   local m, um = minfo:match("(%d+)(%a+)")
   m = m and tonumber(m) or 0
   um = um and um:sub(1, 1):lower() or "b"
   local f = mem_fact[um]
   if f then return m * f else return m end
end

slurm.actions = {
    -- Terminates a command
    -- @param job The job object
    terminate = function(self, job)
        local data, err = job.data.squeue_data
        if not data then
            data, err = squeue(self, job)
            if not data then
                return nil, err
            end
        end
        local state = slurm_to_sga_state[data.JobState]
        if state ~= "FINISHED" then
       self.exec:run(cmds.scancel..job.data.slurmjid)	
           job.data.killed = true
           return true
        else
           return false, "Job is not running"
        end
     end,
 
    -- Gets a command current status
    -- @param job The job object
    -- @return A table with information for each command component (process)
    status = function(self, job)
        -- This function reuses the data periodically fetched by slurm.is_command_done.
        local data, err = job.data.squeue_data
        local processes = {}
        if not data then
            data, err = squeue(self, job)
            if not data then
                return nil, err
            end
        end
        local state = slurm_to_sga_state[data.JobState]
        local walltime, mem, vmem, cput
        if state == "FINISHED" then
            walltime = get_seconds(data.EndTime)
            --mem = get_mem(data.Job.resources_used.mem)
            --vmem = get_mem(data.Job.resources_used.vmem)
            --cput = get_seconds(data.Job.resources_used.cput)
        else
            walltime = get_seconds(data.RunTime)
            mem, vmem, cput = 0, 0, 0
        end
        processes[#processes + 1] = {
            pid = data.JobId:match("^(%d+)"),
            ppid = 0,
            exec_host = data.BatchHost,
            string = job.cmd_string,
            state = state,
            processor_id = "",
            memory_ram_size_mb = mem,
            memory_swap_size_mb = 0,
            cpu_perc = 0,
            cpu_time_sec = cput,
            wall_time_sec = walltime,
            system_time_sec = 0,
            user_time_sec = 0,
            virtual_memory_size_mb = vmem,
            bytes_in_kb = 0,
            bytes_out_kb = 0,
            disk_bytes_read_kb = 0,
            disk_bytes_write_kb = 0,
        }
        for k, v in util.collect_exec_data(job) do
            processes[1][k] = v
            -- copas.step(1)
        end

        return true, { processes = processes }
    end,
}

local function get_node_attributes(ndata)
    local attribs = {}

    for attr, val in ndata:gmatch("(%S+)=(%S+)") do
       attribs[attr] = attribs[attr] or val
    end

    return attribs
end

local function update_node(self, nname, ndata)
    local data = self.nodes_data[nname] or {}
    
    -- static information
    data.clock_mhz = 0 -- FIXME do not return fake information!
    local attribs = get_node_attributes(ndata)
    local np = tonumber(attribs.np) or 1
    data.num_of_cpus = np
    local availmem = tonumber(attribs.availmem) or 0
    data.ram_mb = availmem

    -- dynamic information
    local freemem = attribs.freemem and tonumber(attribs.freemem) or 0
    local totused = availmem - freemem
    local uram
    if totused <= availmem then
       uram = totused
    else
       uram = availmem
    end
    data.ram_used_perc = uram * 100 / availmem
    
    local cpuload = attribs.cpuload and tonumber(attribs.cpuload) or 0
    data.load_avg_1min_perc = cpuload / np
    data.load_avg_5min_perc = data.load_avg_1min_perc  -- FIXME fake!
    data.load_avg_15min_perc = data.load_avg_1min_perc -- FIXME fake!
    -- FIXME: number of jobs and jobs attribute
    self.nodes_data[nname] = data
end
   
function slurm.get_nodes(self)
    local ninfo, err = self.exec:run(cmds.sinfo)
    if not ninfo or ninfo == "" then
       return nil, ("Failed reading data after %s (%s)"):format(cmds.sinfo, err)
    end
    ninfo = "\n"..ninfo -- for uniform matching of node names
    ninfo = ninfo:gsub("\n(%S+) ", "\n<%1> ")  -- <name>   
    for nname, data in ninfo:gmatch("<(%S+)>([^<]+)") do
       nname = nname:gsub("%..*", "") -- remove domain, if exists
       update_node(self, nname, data)
    end
    -- for simplicity, also returns current status data
    return self.nodes_data
 end

---
-- Get nodes current status.
-- @return A table indexed by node name, each entry contains the node´s available resources ("monitoring" data)
function slurm.get_nodes_status(self)
    for nname, _ in pairs(self.nodes_data) do
       local cmd = cmds.sinfo.."-n "..nname
       local ninfo = self.exec:run(cmd)
       if not ninfo then
          self.logger:error("Failed reading data after "..cmd)
       else
          update_node(self, nname, ninfo)
       end
    end
    -- for simplicity, also returns nodes resources
    return self.nodes_data
end

--
-- Gets a local path attributes.
-- @param path_name The path filename
-- @return A table with the file attributes corresponding to path_name, or nil and a message
function slurm.check_path(self, path_name)
    -- FIXME either check path remotely or assume that
    -- the configuration is correct.
    return util.check_local_path(path_name)
end

--
-- Gets a list of child files from the specified local path.
-- @param path_name The local path
-- @return A list of tables with the file attributes corresponding to the children of path_name, or nil and a message
function slurm.list_path(self, path_name)
    -- FIXME either check path remotely or assume that
    -- the configuration is correct.
    return util.list_local_path(path_name)
end

function slurm.new(config, logger)
    local exec_driver, err = exec.init(config)
    if err then
       return nil, err
    end
 
    -- Read optional slurm configuration
    local slurm_config = (config.driver_config and config.driver_config.slurm_config) or {}
    if slurm_config then
       local slurm_config_schema = schema.Record {
          init_dir = schema.Optional(schema.String),
          queue = schema.Optional(schema.String),
          allow_proxy_user = schema.Optional(schema.Boolean),
          proxy_user_group = schema.Optional(schema.String),
       }
       local err = schema.CheckSchema(slurm_config, slurm_config_schema)
       if err then
          return nil, "configuration error in slurm_config: "..tostring(err)
       end
    end
 
    local self = safer.table {
       config = config,
       logger = logger,
       nodes_data = {},
       exec = exec_driver,
       slurm_init_dir = slurm_config.init_dir or nil,
       slurm_queue = slurm_config.queue or nil,
       allow_proxy_user = slurm_config.allow_proxy_user or false,
       proxy_user_group = slurm_config.proxy_user_group or nil,
    }
    return self
end

return slurm

